"""
Pure functions to generate OAuth 2.1 metadata responses.

This module provides generators for OAuth discovery endpoints without any
framework dependencies, making them reusable across different HTTP servers.

Standards Implemented:
    - RFC 8414: OAuth 2.0 Authorization Server Metadata
    - RFC 9728: OAuth 2.0 Protected Resource Metadata
    - OpenID Connect Discovery 1.0

Examples:
    >>> from payments_py.mcp.http import get_oauth_urls
    >>> urls = get_oauth_urls("staging_sandbox")
    >>> print(urls["issuer"])
    'https://nevermined.dev'
"""

from typing import Dict, List, Optional

from ...environments import EnvironmentName, Environments
from ..types.http_types import (
    AuthorizationServerMetadata,
    McpProtectedResourceMetadata,
    OAuthConfig,
    OAuthUrls,
    OidcConfiguration,
    ProtectedResourceMetadata,
    ServerInfoResponse,
)

# =============================================================================
# OAUTH URLS
# =============================================================================


def _build_oauth_urls(frontend_url: str, backend_url: str) -> OAuthUrls:
    """Build OAuth URLs from frontend and backend URLs.

    - issuer and authorizationUri use the frontend (user-facing)
    - tokenUri, jwksUri, userinfoUri use the backend (API)

    Args:
        frontend_url: The frontend URL (e.g., https://nevermined.app).
        backend_url: The backend URL (e.g., https://api.sandbox.nevermined.app).

    Returns:
        OAuth URLs configuration dict.
    """
    # Remove trailing slashes
    frontend = frontend_url.rstrip("/")
    backend = backend_url.rstrip("/")

    return {
        "issuer": frontend,
        "authorizationUri": f"{frontend}/oauth/authorize",
        "tokenUri": f"{backend}/oauth/token",
        "jwksUri": f"{backend}/.well-known/jwks.json",
        "userinfoUri": f"{backend}/oauth/userinfo",
    }


def _get_oauth_urls_for_environment(environment: EnvironmentName) -> OAuthUrls:
    """Get OAuth URLs for an environment.

    Uses frontend and backend URLs from Environments configuration.

    Args:
        environment: The Nevermined environment name.

    Returns:
        OAuth URLs configuration dict.
    """
    env_config = Environments.get(environment, Environments["sandbox"])
    return _build_oauth_urls(env_config.frontend, env_config.backend)


def get_oauth_urls(
    environment: EnvironmentName, overrides: Optional[Dict[str, str]] = None
) -> OAuthUrls:
    """Get OAuth URLs for a given environment with optional overrides.

    Args:
        environment: The Nevermined environment name.
        overrides: Optional dict to override specific URLs.

    Returns:
        Complete OAuth URLs configuration.

    Examples:
        >>> urls = get_oauth_urls("staging_sandbox")
        >>> urls["issuer"]
        'https://nevermined.dev'

        >>> custom_urls = get_oauth_urls("sandbox", {"issuer": "https://custom.com"})
        >>> custom_urls["issuer"]
        'https://custom.com'
    """
    base_urls = _get_oauth_urls_for_environment(environment)
    if overrides:
        base_urls.update(overrides)  # type: ignore
    return base_urls


# =============================================================================
# DEFAULT SCOPES
# =============================================================================

# Default OAuth scopes supported by Nevermined MCP servers
_DEFAULT_SCOPES: List[str] = [
    "openid",
    "profile",
    "credits",
    "mcp:read",
    "mcp:write",
    "mcp:tools",
]


# =============================================================================
# METADATA BUILDERS
# =============================================================================


def build_protected_resource_metadata(config: OAuthConfig) -> ProtectedResourceMetadata:
    """Build Protected Resource Metadata (RFC 9728).

    This metadata tells OAuth clients about the protected resource.

    Args:
        config: OAuth configuration including baseUrl, agentId, and environment.

    Returns:
        Protected Resource Metadata response dict.

    Examples:
        >>> metadata = build_protected_resource_metadata({
        ...     "baseUrl": "http://localhost:5001",
        ...     "agentId": "agent_123",
        ...     "environment": "staging_sandbox"
        ... })
        >>> metadata["resource"]
        'http://localhost:5001'
    """
    scopes = config.get("scopes") or list(_DEFAULT_SCOPES)
    # Get OAuth URLs for validation (not used in this metadata directly)
    _ = get_oauth_urls(config["environment"], config.get("oauthUrls"))

    return {
        "resource": config["baseUrl"],
        "authorization_servers": [config["baseUrl"]],
        "scopes_supported": scopes,
        "bearer_methods_supported": ["header"],
        "resource_documentation": f"{config['baseUrl']}/",
    }


def build_mcp_protected_resource_metadata(
    config: OAuthConfig,
) -> McpProtectedResourceMetadata:
    """Build MCP-specific Protected Resource Metadata.

    Extends the base metadata with MCP capabilities information.

    Args:
        config: OAuth configuration including tools list and protocol version.

    Returns:
        MCP Protected Resource Metadata response dict.

    Examples:
        >>> metadata = build_mcp_protected_resource_metadata({
        ...     "baseUrl": "http://localhost:5001",
        ...     "agentId": "agent_123",
        ...     "environment": "staging_sandbox",
        ...     "tools": ["hello_world", "weather"]
        ... })
        >>> metadata["mcp_capabilities"]["tools"]
        ['hello_world', 'weather']
    """
    scopes = config.get("scopes") or list(_DEFAULT_SCOPES)
    # Get OAuth URLs for validation (not used in this metadata directly)
    _ = get_oauth_urls(config["environment"], config.get("oauthUrls"))

    return {
        "resource": f"{config['baseUrl']}/mcp",
        "authorization_servers": [config["baseUrl"]],
        "scopes_supported": scopes,
        "scopes_required": scopes,
        "bearer_methods_supported": ["header"],
        "resource_documentation": f"{config['baseUrl']}/",
        "mcp_capabilities": {
            "tools": config.get("tools") or [],
            "protocol_version": config.get("protocolVersion") or "2024-11-05",
        },
    }


def build_authorization_server_metadata(
    config: OAuthConfig,
) -> AuthorizationServerMetadata:
    """Build OAuth Authorization Server Metadata (RFC 8414).

    This metadata describes the OAuth authorization server configuration.

    Args:
        config: OAuth configuration.

    Returns:
        Authorization Server Metadata response dict.

    Examples:
        >>> metadata = build_authorization_server_metadata({
        ...     "baseUrl": "http://localhost:5001",
        ...     "agentId": "agent_123",
        ...     "environment": "staging_sandbox"
        ... })
        >>> metadata["issuer"]
        'https://nevermined.dev'
    """
    oauth_urls = get_oauth_urls(config["environment"], config.get("oauthUrls"))
    scopes = config.get("scopes") or list(_DEFAULT_SCOPES)

    return {
        "issuer": oauth_urls["issuer"],
        "authorization_endpoint": oauth_urls["authorizationUri"],
        "token_endpoint": oauth_urls["tokenUri"],
        "registration_endpoint": f"{config['baseUrl']}/register",
        "jwks_uri": oauth_urls["jwksUri"],
        "response_types_supported": ["code"],
        "grant_types_supported": ["authorization_code", "refresh_token"],
        "code_challenge_methods_supported": ["S256"],
        "scopes_supported": scopes,
        "token_endpoint_auth_methods_supported": ["client_secret_post"],
        "subject_types_supported": ["public"],
    }


def build_oidc_configuration(config: OAuthConfig) -> OidcConfiguration:
    """Build OpenID Connect Discovery Metadata.

    Provides OIDC-compatible configuration for clients that expect OpenID Connect.

    Args:
        config: OAuth configuration.

    Returns:
        OIDC Configuration response dict.

    Examples:
        >>> metadata = build_oidc_configuration({
        ...     "baseUrl": "http://localhost:5001",
        ...     "agentId": "agent_123",
        ...     "environment": "staging_sandbox"
        ... })
        >>> "openid" in metadata["scopes_supported"]
        True
    """
    oauth_urls = get_oauth_urls(config["environment"], config.get("oauthUrls"))
    scopes = config.get("scopes") or list(_DEFAULT_SCOPES)
    all_scopes = scopes if "openid" in scopes else ["openid"] + scopes

    return {
        "issuer": oauth_urls["issuer"],
        "authorization_endpoint": oauth_urls["authorizationUri"],
        "token_endpoint": oauth_urls["tokenUri"],
        "jwks_uri": oauth_urls["jwksUri"],
        "userinfo_endpoint": oauth_urls["userinfoUri"],
        "registration_endpoint": f"{config['baseUrl']}/register",
        "response_types_supported": ["code"],
        "grant_types_supported": ["authorization_code", "refresh_token"],
        "code_challenge_methods_supported": ["S256"],
        "token_endpoint_auth_methods_supported": ["none", "client_secret_post"],
        "subject_types_supported": ["public"],
        "id_token_signing_alg_values_supported": ["RS256", "HS256"],
        "scopes_supported": all_scopes,
        "claims_supported": ["sub", "iss", "aud", "exp", "iat", "name", "email"],
    }


def build_server_info_response(
    config: OAuthConfig,
    version: Optional[str] = None,
    description: Optional[str] = None,
) -> ServerInfoResponse:
    """Build server info response for the root endpoint.

    Args:
        config: OAuth configuration.
        version: Server version. Defaults to '1.0.0'.
        description: Server description.

    Returns:
        Server info response dict.

    Examples:
        >>> info = build_server_info_response({
        ...     "baseUrl": "http://localhost:5001",
        ...     "agentId": "abc123",
        ...     "environment": "staging_sandbox",
        ...     "serverName": "my-mcp",
        ...     "tools": ["hello"]
        ... }, version="0.1.0")
        >>> info["name"]
        'my-mcp'
        >>> info["tools"]
        ['hello']
    """
    oauth_urls = get_oauth_urls(config["environment"], config.get("oauthUrls"))
    scopes = config.get("scopes") or list(_DEFAULT_SCOPES)

    return {
        "name": config.get("serverName") or "MCP Server",
        "version": version or "1.0.0",
        "description": description
        or "MCP server with Nevermined OAuth integration via Streamable HTTP",
        "endpoints": {
            "mcp": f"{config['baseUrl']}/mcp",
            "health": f"{config['baseUrl']}/health",
            "register": f"{config['baseUrl']}/register",
        },
        "oauth": {
            "authorization_server_metadata": f"{config['baseUrl']}/.well-known/oauth-authorization-server",
            "protected_resource_metadata": f"{config['baseUrl']}/.well-known/oauth-protected-resource",
            "openid_configuration": f"{config['baseUrl']}/.well-known/openid-configuration",
            "authorization_endpoint": oauth_urls["authorizationUri"],
            "token_endpoint": oauth_urls["tokenUri"],
            "jwks_uri": oauth_urls["jwksUri"],
            "registration_endpoint": f"{config['baseUrl']}/register",
            "client_id": config["agentId"],
            "scopes": scopes,
        },
        "tools": config.get("tools") or [],
        "resources": config.get("resources") or [],
        "prompts": config.get("prompts") or [],
    }



"""
observability_api.py
Provides reusable utilities for wrapping API calls with Helicone logging for AI agents
"""

import time
import math
import json
from typing import Dict, Any, Optional, Callable, Awaitable, TypeVar, Generic, Union
from dataclasses import dataclass
from urllib.parse import urljoin
from helicone_helpers import HeliconeManualLogger
from helicone_helpers.manual_logger import HeliconeResultRecorder

from payments_py.api.base_payments import BasePaymentsAPI
from payments_py.common.types import PaymentOptions, StartAgentRequest

# Type variables for generic functions
T = TypeVar("T")
R = TypeVar("R")

CustomProperties = Dict[str, str]


@dataclass
class HeliconePayloadConfig:
    """Configuration for creating a Helicone payload"""

    model: str
    input_data: Dict[str, Any]
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None
    n: Optional[int] = None
    stream: Optional[bool] = None


@dataclass
class UsageDetails:
    """Token usage details"""

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int
    prompt_tokens_details: Optional[Dict[str, int]] = None
    completion_tokens_details: Optional[Dict[str, int]] = None


@dataclass
class HeliconeResponseConfig:
    """Configuration for creating a Helicone response"""

    id_prefix: str
    model: str
    result_data: Any
    usage: UsageDetails
    system_fingerprint: Optional[str] = None


@dataclass
class NeverminedHeliconeHeaders:
    """Nevermined-specific Helicone headers"""

    helicone_auth: str
    account_address: str
    consumer_address: str
    agent_id: str
    plan_id: str
    plan_type: str
    plan_name: str
    agent_name: str
    agent_request_id: str
    price_per_credit: str
    environment_name: str
    batch: str
    is_margin_based: str
    margin_percent: str


@dataclass
class ChatOpenAIConfiguration:
    """Configuration for ChatOpenAI with Helicone"""

    model: str
    api_key: str
    configuration: Dict[str, Any]


@dataclass
class OpenAIConfiguration:
    """Configuration for OpenAI client with Helicone"""

    api_key: str
    base_url: str
    default_headers: Dict[str, str]


def get_default_helicone_headers(
    helicone_api_key: str,
    account_address: str,
    environment_name: str,
    start_agent_request: StartAgentRequest,
    custom_properties: CustomProperties,
) -> Dict[str, str]:
    """Build default Helicone headers from StartAgentRequest object"""

    # Extract values from StartAgentRequest object
    consumer_address = start_agent_request.balance.holder_address
    agent_id = start_agent_request.agent_id
    plan_id = start_agent_request.balance.plan_id
    plan_type = start_agent_request.balance.plan_type
    plan_name = start_agent_request.balance.plan_name
    agent_name = start_agent_request.agent_name
    agent_request_id = start_agent_request.agent_request_id
    price_per_credit = start_agent_request.balance.price_per_credit
    batch = start_agent_request.batch

    # Build Nevermined headers
    nevermined_headers = {
        "Helicone-Auth": f"Bearer {helicone_api_key}",
        "Helicone-Property-accountAddress": account_address,
        "Helicone-Property-consumerAddress": consumer_address,
        "Helicone-Property-agentId": agent_id,
        "Helicone-Property-planId": plan_id,
        "Helicone-Property-planType": plan_type,
        "Helicone-Property-planName": plan_name,
        "Helicone-Property-agentName": agent_name,
        "Helicone-Property-agentRequestId": agent_request_id,
        "Helicone-Property-pricePerCredit": str(price_per_credit),
        "Helicone-Property-environmentName": environment_name,
        "Helicone-Property-batch": str(batch).lower(),
        "Helicone-Property-ismarginBased": "false",
        "Helicone-Property-marginPercent": "0",
    }

    # Add custom property headers
    custom_headers = {}
    for key, value in custom_properties.items():
        custom_headers[f"Helicone-Property-{key}"] = str(value)

    return {**nevermined_headers, **custom_headers}


def create_helicone_payload(config: HeliconePayloadConfig) -> Dict[str, Any]:
    """Creates a standardized Helicone payload for API logging"""
    return {
        "model": config.model,
        "temperature": config.temperature or 1.0,
        "top_p": config.top_p or 1.0,
        "frequency_penalty": config.frequency_penalty or 0.0,
        "presence_penalty": config.presence_penalty or 0.0,
        "n": config.n or 1,
        "stream": config.stream or False,
        "messages": [
            {
                "role": "user",
                "content": json.dumps(config.input_data),
            }
        ],
    }


def create_helicone_response(config: HeliconeResponseConfig) -> Dict[str, Any]:
    """Creates a standardized Helicone response for API logging"""
    timestamp = int(time.time() * 1000)

    return {
        "id": f"{config.id_prefix}-{timestamp}",
        "object": "chat.completion",
        "created": int(timestamp / 1000),
        "model": config.model,
        "choices": [
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": json.dumps(config.result_data),
                    "refusal": None,
                    "annotations": [],
                },
                "logprobs": None,
                "finish_reason": "stop",
            }
        ],
        "usage": {
            "prompt_tokens": config.usage.prompt_tokens,
            "completion_tokens": config.usage.completion_tokens,
            "total_tokens": config.usage.total_tokens,
            "prompt_tokens_details": config.usage.prompt_tokens_details
            or {
                "cached_tokens": 0,
                "audio_tokens": 0,
            },
            "completion_tokens_details": config.usage.completion_tokens_details
            or {
                "reasoning_tokens": 0,
                "audio_tokens": 0,
                "accepted_prediction_tokens": 0,
                "rejected_prediction_tokens": 0,
            },
        },
        "service_tier": "default",
        "system_fingerprint": config.system_fingerprint or f"fp_{timestamp}",
    }


async def with_manual_logging(
    agent_name: str,
    payload_config: HeliconePayloadConfig,
    operation: Callable[[], Awaitable[T]],
    result_extractor: Callable[[T], R],
    usage_calculator: Callable[[T], UsageDetails],
    response_id_prefix: str,
    helicone_api_key: str,
    helicone_manual_logging_url: str,
    account_address: str,
    environment_name: str,
    start_agent_request: StartAgentRequest,
    custom_properties: CustomProperties,
) -> R:
    """
    Wraps an async operation with manual logging
    """
    # Get default headers
    default_headers = get_default_helicone_headers(
        helicone_api_key,
        account_address,
        environment_name,
        start_agent_request,
        custom_properties,
    )

    # Initialize the official Helicone manual logger
    helicone_logger = HeliconeManualLogger(
        api_key=helicone_api_key,
        logging_endpoint=helicone_manual_logging_url,
        headers=default_headers,
    )

    helicone_payload = create_helicone_payload(payload_config)

    # Use the official Helicone manual logger pattern
    async def logging_operation(result_recorder: HeliconeResultRecorder) -> R:
        internal_result = await operation()
        usage = usage_calculator(internal_result)
        extracted_result = result_extractor(internal_result)

        helicone_response = create_helicone_response(
            HeliconeResponseConfig(
                id_prefix=response_id_prefix,
                model=payload_config.model,
                result_data=extracted_result,
                usage=usage,
                system_fingerprint=getattr(extracted_result, "job_id", None)
                and f'fp_{getattr(extracted_result, "job_id", "")}',
            )
        )

        result_recorder.append_results(helicone_response)
        return extracted_result

    return await helicone_logger.log_request(helicone_payload, logging_operation)


def calculate_image_usage(pixels: int) -> UsageDetails:
    """Helper function to calculate usage for image operations based on pixels"""
    return UsageDetails(
        prompt_tokens=0,
        completion_tokens=pixels,
        total_tokens=pixels,
        prompt_tokens_details={"cached_tokens": 0, "audio_tokens": 0},
        completion_tokens_details={
            "reasoning_tokens": 0,
            "audio_tokens": 0,
            "accepted_prediction_tokens": 0,
            "rejected_prediction_tokens": 0,
        },
    )


def calculate_video_usage() -> UsageDetails:
    """Helper function to calculate usage for video operations (typically 1 token)"""
    return UsageDetails(
        prompt_tokens=0,
        completion_tokens=1,
        total_tokens=1,
        prompt_tokens_details={"cached_tokens": 0, "audio_tokens": 0},
        completion_tokens_details={
            "reasoning_tokens": 0,
            "audio_tokens": 0,
            "accepted_prediction_tokens": 0,
            "rejected_prediction_tokens": 0,
        },
    )


def calculate_song_usage(tokens: int) -> UsageDetails:
    """Helper function to calculate usage for song operations based on tokens/quota"""
    return UsageDetails(
        prompt_tokens=0,
        completion_tokens=tokens,
        total_tokens=tokens,
        prompt_tokens_details={"cached_tokens": 0, "audio_tokens": 0},
        completion_tokens_details={
            "reasoning_tokens": 0,
            "audio_tokens": 0,
            "accepted_prediction_tokens": 0,
            "rejected_prediction_tokens": 0,
        },
    )


def calculate_dummy_song_usage() -> UsageDetails:
    """Helper function to calculate usage for dummy song operations"""
    return calculate_song_usage(6)  # Default dummy token count


def with_langchain(
    model: str,
    api_key: str,
    helicone_api_key: str,
    helicone_base_logging_url: str,
    account_address: str,
    environment_name: str,
    start_agent_request: StartAgentRequest,
    custom_properties: CustomProperties,
) -> ChatOpenAIConfiguration:
    """
    Creates a ChatOpenAI configuration with logging enabled
    """
    default_headers = get_default_helicone_headers(
        helicone_api_key,
        account_address,
        environment_name,
        start_agent_request,
        custom_properties,
    )

    return ChatOpenAIConfiguration(
        model=model,
        api_key=api_key,
        configuration={
            "base_url": helicone_base_logging_url,
            "default_headers": default_headers,
        },
    )


def with_openai(
    api_key: str,
    helicone_api_key: str,
    helicone_base_logging_url: str,
    account_address: str,
    environment_name: str,
    start_agent_request: StartAgentRequest,
    custom_properties: CustomProperties,
) -> OpenAIConfiguration:
    """
    Creates an OpenAI client configuration with logging enabled
    """
    default_headers = get_default_helicone_headers(
        helicone_api_key,
        account_address,
        environment_name,
        start_agent_request,
        custom_properties,
    )

    return OpenAIConfiguration(
        api_key=api_key,
        base_url=helicone_base_logging_url,
        default_headers=default_headers,
    )


class ObservabilityAPI(BasePaymentsAPI):
    """
    The ObservabilityAPI class provides methods to wrap API calls with Helicone logging
    """

    def __init__(self, options: PaymentOptions):
        super().__init__(options)

        # TODO: For testing purposes only. Remove once helicone is deployed to staging
        # Get Helicone API key from environment variable and override the base class property
        import os

        self.helicone_api_key = os.getenv("HELICONE_API_KEY") or self.helicone_api_key

        self.helicone_base_logging_url = urljoin(
            self.environment.helicone_url, "jawn/v1/gateway/oai/v1"
        )
        self.helicone_manual_logging_url = urljoin(
            self.environment.helicone_url, "jawn/v1/trace/custom/v1/log"
        )

    @classmethod
    def get_instance(cls, options: PaymentOptions) -> "ObservabilityAPI":
        """
        This method is used to create a singleton instance of the ObservabilityAPI class.
        """
        return cls(options)

    async def with_manual_logging(
        self,
        agent_name: str,
        payload_config: HeliconePayloadConfig,
        operation: Callable[[], Awaitable[T]],
        result_extractor: Callable[[T], R],
        usage_calculator: Callable[[T], UsageDetails],
        response_id_prefix: str,
        start_agent_request: StartAgentRequest,
        custom_properties: CustomProperties,
    ) -> R:
        """
        Wraps an async operation with manual logging
        """
        return await with_manual_logging(
            agent_name,
            payload_config,
            operation,
            result_extractor,
            usage_calculator,
            response_id_prefix,
            self.helicone_api_key,
            self.helicone_manual_logging_url,
            self.account_address,
            self.environment_name,
            start_agent_request,
            custom_properties,
        )

    def with_langchain(
        self,
        model: str,
        api_key: str,
        start_agent_request: StartAgentRequest,
        custom_properties: CustomProperties,
    ) -> ChatOpenAIConfiguration:
        """
        Creates a ChatOpenAI configuration with logging enabled
        """
        return with_langchain(
            model,
            api_key,
            self.helicone_api_key,
            self.helicone_base_logging_url,
            self.account_address,
            self.environment_name,
            start_agent_request,
            custom_properties,
        )

    def with_openai(
        self,
        api_key: str,
        start_agent_request: StartAgentRequest,
        custom_properties: CustomProperties,
    ) -> OpenAIConfiguration:
        """
        Creates an OpenAI client configuration with logging enabled
        """
        return with_openai(
            api_key,
            self.helicone_api_key,
            self.helicone_base_logging_url,
            self.account_address,
            self.environment_name,
            start_agent_request,
            custom_properties,
        )

    def calculate_image_usage(self, pixels: int) -> UsageDetails:
        """Helper function to calculate usage for image operations based on pixels"""
        return calculate_image_usage(pixels)

    def calculate_video_usage(self) -> UsageDetails:
        """Helper function to calculate usage for video operations (typically 1 token)"""
        return calculate_video_usage()

    def calculate_song_usage(self, tokens: int) -> UsageDetails:
        """Helper function to calculate usage for song operations based on tokens/quota"""
        return calculate_song_usage(tokens)

    def calculate_dummy_song_usage(self) -> UsageDetails:
        """Helper function to calculate usage for dummy song operations"""
        return calculate_dummy_song_usage()

    def create_helicone_payload(self, config: HeliconePayloadConfig) -> Dict[str, Any]:
        """Creates a standardized Helicone payload for API logging"""
        return create_helicone_payload(config)

    def create_helicone_response(
        self, config: HeliconeResponseConfig
    ) -> Dict[str, Any]:
        """Creates a standardized Helicone response for API logging"""
        return create_helicone_response(config)



"""
Helpers to build MCP-compatible `extra` objects from HTTP requests/headers.
"""

from typing import Any, Dict


def build_extra_from_http_headers(headers: Dict[str, Any]) -> Dict[str, Any]:
    """Build an MCP ``extra`` object from raw HTTP headers mapping.

    Args:
        headers: Dictionary of HTTP headers.

    Returns:
        Extra object compatible with MCP libraries.
    """
    return {"requestInfo": {"headers": headers or {}}}


def build_extra_from_http_request(req: Any) -> Dict[str, Any]:
    """Build an MCP ``extra`` object from a request-like object.

    Args:
        req: Object or dict that may carry a ``headers`` attribute or key.

    Returns:
        Extra object compatible with MCP libraries.
    """
    headers = getattr(req, "headers", None)
    if headers is None and isinstance(req, dict):
        headers = req.get("headers")
    return build_extra_from_http_headers(headers or {})


def build_extra_from_fastmcp_context(ctx: Any) -> Dict[str, Any]:
    """
    Create an MCP ``extra`` object from a FastMCP ``Context``.

    Parameters
    ----------
    ctx : Any
        FastMCP context instance passed to tool/resource/prompt functions.

    Returns
    -------
    dict
        Extra object including request headers if present.
    """
    try:
        if ctx and getattr(ctx, "request_context", None):
            # Prefer direct request attr; fallback to meta.request_context
            req = getattr(ctx.request_context, "request", None)
            if req is None and getattr(ctx.request_context, "meta", None) is not None:
                req = getattr(ctx.request_context.meta, "request_context", None)
            if req is None:
                req = getattr(ctx.request_context, "meta", None)
                req = getattr(req, "request", None) if req is not None else None
            if req is None:
                return build_extra_from_http_headers({})
            req_headers = getattr(req, "headers", {}) or {}
            # Starlette's Headers object is a Mapping; convert safely
            if hasattr(req_headers, "items"):
                headers = {str(k): str(v) for k, v in req_headers.items()}
            elif hasattr(req_headers, "raw"):
                headers = {k.decode(): v.decode() for k, v in (req_headers.raw or [])}
            elif isinstance(req_headers, dict):
                headers = {str(k): str(v) for k, v in req_headers.items()}
            else:
                headers = {}
            return build_extra_from_http_headers(headers)
    except Exception:
        pass
    return build_extra_from_http_headers({})



"""
Constants for AgentCore x402 Integration.
"""

# x402 HTTP Transport header names (v2 spec)
# @see https://github.com/coinbase/x402/blob/main/specs/transports-v2/http.md
X402_HEADERS = {
    # Client sends payment token in this header
    "PAYMENT_SIGNATURE": "payment-signature",
    # Server sends PaymentRequired in this header (base64-encoded)
    "PAYMENT_REQUIRED": "payment-required",
    # Server sends settlement receipt in this header (base64-encoded)
    "PAYMENT_RESPONSE": "payment-response",
    # Internal: passes agent_request_id from REQUEST to RESPONSE phase
    "AGENT_REQUEST_ID": "x-nvm-agent-request-id",
}

# MCP JSON-RPC 2.0 methods
MCP_METHODS = {
    "TOOLS_CALL": "tools/call",
    "TOOLS_LIST": "tools/list",
    "RESOURCES_LIST": "resources/list",
    "RESOURCES_READ": "resources/read",
    "PROMPTS_LIST": "prompts/list",
    "PROMPTS_GET": "prompts/get",
}

# Default billable methods
DEFAULT_BILLABLE_METHODS = [MCP_METHODS["TOOLS_CALL"]]

# Interceptor output version
INTERCEPTOR_OUTPUT_VERSION = "1.0"



"""
Main Payments class for the Nevermined Payments protocol.
"""

from typing import Dict, Any
from payments_py.common.payments_error import PaymentsError
from payments_py.common.types import PaymentOptions
from payments_py.api.query_api import AIQueryApi
from payments_py.api.base_payments import BasePaymentsAPI
from payments_py.api.plans_api import PlansAPI
from payments_py.api.agents_api import AgentsAPI
from payments_py.api.requests_api import AgentRequestsAPI
from payments_py.api.observability_api import ObservabilityAPI
from payments_py.api.contracts_api import ContractsAPI
from payments_py.x402.facilitator_api import FacilitatorAPI
from payments_py.x402.token import X402TokenAPI

# A2A integration
from payments_py.a2a.agent_card import build_payment_agent_card
from payments_py.a2a.client_registry import ClientRegistry


class Payments(BasePaymentsAPI):
    """
    Main class that interacts with the Nevermined payments API.
    Use `Payments.get_instance` for server-side usage.

    The library provides methods to manage AI Agents, Plans & process AI Agent Requests.

    Each of these functionalities is encapsulated in its own API class:
    - `plans`: Manages AI Plans, including registration and ordering and retrieving plan details.
    - `agents`: Handles AI Agents, including registration of AI Agents and access token generation.
    - `requests`: Manages requests received by AI Agents, including validation and tracking.
    - `facilitator`: Handles X402 permission verification and settlement for AI Agents acting as facilitators.
    """

    def __init__(self, options: PaymentOptions, is_browser_instance: bool = False):
        """
        Initialize the Payments class.

        Args:
            options: The initialization options
            is_browser_instance: Whether this is a browser instance (default False)
        """
        super().__init__(options)
        self.is_browser_instance = is_browser_instance
        self._initialize_api(options)

    @classmethod
    def get_instance(cls, options: PaymentOptions) -> "Payments":
        """
        Get an instance of the Payments class for server-side usage.

        Args:
            options: The options to initialize the payments class

        Returns:
            An instance of Payments

        Raises:
            PaymentsError: If nvm_api_key is missing
        """
        if not options.nvm_api_key:
            raise PaymentsError.unauthorized("Nevermined API Key is required")
        return cls(options, False)

    @classmethod
    def get_browser_instance(cls, options: PaymentOptions) -> "Payments":
        """
        Get an instance of the Payments class for browser usage.

        Args:
            options: The options to initialize the payments class

        Returns:
            An instance of Payments

        Raises:
            PaymentsError: If return_url is missing
        """
        if not options.return_url:
            raise PaymentsError.validation("return_url is required")
        return cls(options, True)

    def _initialize_api(self, options: PaymentOptions) -> None:
        """
        Initialize the AI Query Protocol API.
        """
        self.plans = PlansAPI.get_instance(options)
        self.agents = AgentsAPI.get_instance(options)
        self.requests = AgentRequestsAPI.get_instance(options)
        self.query = AIQueryApi.get_instance()
        self.observability = ObservabilityAPI.get_instance(options)
        self.facilitator = FacilitatorAPI.get_instance(options)
        self.x402 = X402TokenAPI.get_instance(options)
        self.contracts_api = ContractsAPI(options)

        # Cached MCP integration
        self._mcp_integration = None

        # Cached AgentCore integration
        self._agentcore_api = None

    @property
    def contracts(self):
        """
        Get contract addresses from the deployment info endpoint.

        Returns:
            Contracts model with all contract addresses accessible via snake_case properties

        Example::
            template_address = payments.contracts.pay_as_you_go_template
        """
        return self.contracts_api.contracts

    @property
    def mcp(self):
        """
        Returns the MCP integration API. The instance is memoized so that configuration
        set via configure({ agentId, serverName }) persists across calls.
        """
        if self._mcp_integration is None:
            # Local import to avoid import cycles
            from payments_py.mcp.index import build_mcp_integration  # noqa: WPS433

            self._mcp_integration = build_mcp_integration(self)
        return self._mcp_integration

    @property
    def agentcore(self):
        """
        Returns the AgentCore integration API for AWS Bedrock.

        Provides methods to create x402 payment interceptors for
        AgentCore Gateway Lambda functions.

        Example:
            ```python
            interceptor = payments.agentcore.create_interceptor(plan_id="123")
            handler = payments.agentcore.create_lambda_handler(plan_id="123")
            ```
        """
        if self._agentcore_api is None:
            # Local import to avoid import cycles
            from payments_py.x402.agentcore import AgentCoreAPI  # noqa: WPS433

            self._agentcore_api = AgentCoreAPI(self)
        return self._agentcore_api

    @property
    def is_logged_in(self) -> bool:
        """
        Check if a user is logged in.

        Returns:
            True if the user is logged in

        Note: This is a browser-only function.
        """
        raise PaymentsError.internal("This is a browser-only function")
        return bool(self.nvm_api_key)

    def connect(self) -> None:
        """
        Initiates the connect flow. The user's browser will be redirected to
        the Nevermined App login page.

        Note: This is a browser-only function.
        """
        raise PaymentsError.internal("This is a browser-only function")
        url = f"{self.environment.frontend}/login?returnUrl={self.return_url}"
        # In a browser environment, this would redirect
        print(f"Redirecting to: {url}")

    def logout(self) -> None:
        """
        Logs out the user by removing the NVM API key.
        """
        raise PaymentsError.internal("This is a browser-only function")
        self.nvm_api_key = ""

        # ---------------------------------------------------------------------
        # A2A integration helpers
        # ---------------------------------------------------------------------

    # A2A client registry cache
    _a2a_registry: "ClientRegistry | None" = None  # type: ignore[name-defined]

    @property
    def a2a(self) -> Dict[str, Any]:
        from payments_py.a2a.server import PaymentsA2AServer

        """Expose A2A helpers (start server / get client) for Nevermined Payments."""
        # Local imports to avoid circular dependencies

        def _get_client(**options: Any):  # noqa: WPS430
            if self._a2a_registry is None:
                self._a2a_registry = ClientRegistry(self)  # type: ignore[arg-type]
            return self._a2a_registry.get_client(**options)

        return {
            # type: ignore[lambda-assign]
            "start": lambda **opts: PaymentsA2AServer.start(
                payments_service=self, **opts
            ),
            "get_client": _get_client,
        }

    a2a_helpers: Dict[str, Any] = {  # noqa: WPS110
        "build_payment_agent_card": build_payment_agent_card,
    }



"""Common types and utilities for the Nevermined Payments Python SDK."""

from payments_py.common.types import (
    PaymentOptions,
    AgentMetadata,
    AgentAPIAttributes,
    AuthType,
    PlanMetadata,
    PlanPriceConfig,
    PlanCreditsConfig,
    PlanBalance,
    PlanPriceType,
    PlanCreditsType,
    PlanRedemptionType,
    AgentTaskStatus,
    PaginationOptions,
    TrackAgentSubTaskDto,
    StartAgentRequest,
    AgentAccessCredentials,
    NvmAPIResult,
)
from payments_py.common.payments_error import PaymentsError
from payments_py.utils import (
    get_ai_hub_open_api_url,
    get_service_host_from_endpoints,
)

__all__ = [
    "PaymentOptions",
    "AgentMetadata",
    "AgentAPIAttributes",
    "AuthType",
    "PlanMetadata",
    "PlanPriceConfig",
    "PlanCreditsConfig",
    "PlanBalance",
    "PlanPriceType",
    "PlanCreditsType",
    "PlanRedemptionType",
    "AgentTaskStatus",
    "PaginationOptions",
    "TrackAgentSubTaskDto",
    "StartAgentRequest",
    "AgentAccessCredentials",
    "NvmAPIResult",
    "PaymentsError",
    "get_ai_hub_open_api_url",
    "get_service_host_from_endpoints",
]



"""
Nevermined Extension Types

Defines the structure and types for the Nevermined payment extension.
"""

from typing import Optional
from typing_extensions import TypedDict

# Extension identifier constants
# Base identifier for Nevermined extensions
NEVERMINED = "nevermined"


# Qualified extension identifiers for multiple plans
# Format: "nevermined:{plan_type}" where plan_type is a descriptive identifier
# Examples: "nevermined:credits", "nevermined:payasyougo"
def nevermined_extension_key(plan_type: str) -> str:
    """
    Generate a qualified extension key for a Nevermined plan.

    Args:
        plan_type: Descriptive identifier for the plan (e.g., "credits", "payasyougo")

    Returns:
        Qualified extension key (e.g., "nevermined:credits")
    """
    return f"{NEVERMINED}:{plan_type}"


class NeverminedInfo(TypedDict, total=False):
    """
    Information for Nevermined payments.

    This is the 'info' part of the Nevermined extension, containing
    the actual payment requirements data.

    Required fields:
        plan_id: Nevermined pricing plan ID
        agent_id: Nevermined AI agent ID
        max_amount: Maximum credits to burn per request (as string)
        network: Blockchain network (e.g., "base-sepolia")
        scheme: Payment scheme (e.g., "contract")

    Optional fields:
        environment: Nevermined environment ("sandbox", "live")
        subscriber_address: Subscriber's blockchain address
    """

    # Required fields
    plan_id: str
    agent_id: str
    max_amount: str
    network: str
    scheme: str

    # Optional fields
    environment: Optional[str]
    subscriber_address: Optional[str]


class NeverminedExtension(TypedDict):
    """
    Complete Nevermined extension structure (info + schema).

    Follows the x402 v2 extension pattern where extensions contain:
    - info: The actual extension data
    - schema: JSON Schema validating the info

    This structure allows for self-validating, machine-readable metadata.
    """

    info: NeverminedInfo
    schema: dict


__all__ = [
    "NEVERMINED",
    "nevermined_extension_key",
    "NeverminedInfo",
    "NeverminedExtension",
]



"""
PaymentsMCP - Decorator-based MCP server with Nevermined payments and OAuth.

This module provides a decorator-based API similar to FastMCP but with integrated:
- OAuth 2.1 authentication
- Credit redemption for monetizing tools, resources, and prompts
- Full Nevermined payments integration

Example:
    >>> from payments_py import Payments
    >>> from payments_py.mcp import PaymentsMCP
    >>>
    >>> payments = Payments(nvm_api_key="...", environment="staging_sandbox")
    >>> mcp = PaymentsMCP(payments, name="my-server", agent_id="did:nv:...")
    >>>
    >>> @mcp.tool(credits=5)
    >>> def weather(city: str) -> str:
    ...     return f"Weather in {city}: 22°C"
    >>>
    >>> @mcp.resource("data://config", credits=2)
    >>> def get_config() -> str:
    ...     return '{"version": "1.0.0"}'
    >>>
    >>> @mcp.prompt(credits=1)
    >>> def greeting(name: str) -> list:
    ...     return [{"role": "user", "content": f"Hello {name}!"}]
    >>>
    >>> # Run the server with OAuth
    >>> await mcp.start(port=3002)
"""

import asyncio
import functools
import inspect
import json
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, get_type_hints

from .core.server_manager import McpServerManager, create_server_manager
from .types import CreditsOption

# Type for any function
F = TypeVar("F", bound=Callable[..., Any])


def _extract_input_schema(fn: Callable) -> Dict[str, Any]:
    """Extract JSON Schema from function type hints.

    Args:
        fn: Function to extract schema from.

    Returns:
        JSON Schema dict for the function parameters.
    """
    schema: Dict[str, Any] = {
        "type": "object",
        "properties": {},
        "required": [],
    }

    try:
        hints = get_type_hints(fn)
    except Exception:
        hints = {}

    sig = inspect.signature(fn)

    for param_name, param in sig.parameters.items():
        # Skip self, cls, and special parameters
        if param_name in ("self", "cls", "context", "ctx", "extra"):
            continue

        # Determine type
        param_type = hints.get(param_name, param.annotation)
        json_type = "string"  # default

        if param_type != inspect.Parameter.empty:
            if param_type == int:
                json_type = "integer"
            elif param_type == float:
                json_type = "number"
            elif param_type == bool:
                json_type = "boolean"
            elif param_type == list or (
                hasattr(param_type, "__origin__") and param_type.__origin__ == list
            ):
                json_type = "array"
            elif param_type == dict or (
                hasattr(param_type, "__origin__") and param_type.__origin__ == dict
            ):
                json_type = "object"

        prop: Dict[str, Any] = {"type": json_type}

        # Add description from docstring if available
        if fn.__doc__:
            # Simple extraction - look for ":param name:" pattern
            import re

            match = re.search(
                rf":param\s+{param_name}:\s*(.+?)(?=:param|:return|$)",
                fn.__doc__,
                re.DOTALL,
            )
            if match:
                prop["description"] = match.group(1).strip()

        schema["properties"][param_name] = prop

        # Add to required if no default
        if param.default == inspect.Parameter.empty:
            schema["required"].append(param_name)

    return schema


class PaymentsMCP:
    """MCP server with integrated Nevermined payments and OAuth.

    This class provides a decorator-based API similar to FastMCP but uses
    McpServerManager internally, which includes:
    - OAuth 2.1 authentication endpoints
    - Client registration
    - Credit redemption (paywall)
    - Session management
    - Full MCP protocol support

    Attributes:
        payments: The Payments instance for authentication and credit redemption.
        name: Server name.
        agent_id: Nevermined agent DID.

    Example:
        >>> payments = Payments(nvm_api_key="...", environment="staging_sandbox")
        >>> mcp = PaymentsMCP(payments, name="calculator", agent_id="did:nv:...")
        >>>
        >>> @mcp.tool(credits=1)
        >>> def add(a: int, b: int) -> int:
        ...     return a + b
        >>>
        >>> # Dynamic credits based on result
        >>> @mcp.tool(credits=lambda args, result: len(result) // 100)
        >>> def generate_text(prompt: str) -> str:
        ...     return "Generated text..."
        >>>
        >>> await mcp.start(port=3002)
    """

    def __init__(
        self,
        payments: Any,
        name: str = "payments-mcp-server",
        agent_id: Optional[str] = None,
        version: str = "1.0.0",
        description: Optional[str] = None,
    ) -> None:
        """Initialize PaymentsMCP.

        Args:
            payments: Initialized Payments instance.
            name: Server name (default: "payments-mcp-server").
            agent_id: Nevermined agent DID for authentication.
            version: Server version (default: "1.0.0").
            description: Server description.
        """
        self.payments = payments
        self.name = name
        self.agent_id = agent_id
        self.version = version
        self.description = description

        # Use McpServerManager for OAuth and paywall support
        self._manager: McpServerManager = create_server_manager(payments)

        # Track registered handlers for introspection
        self._registered_tools: Dict[str, Dict[str, Any]] = {}
        self._registered_resources: Dict[str, Dict[str, Any]] = {}
        self._registered_prompts: Dict[str, Dict[str, Any]] = {}

    def tool(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        credits: Optional[CreditsOption] = None,
        on_redeem_error: str = "ignore",
    ) -> Callable[[F], F]:
        """Decorator to register a tool with optional credit redemption.

        The decorated function's type hints are used to generate the input schema.
        OAuth authentication and credit redemption are handled automatically.

        Args:
            name: Optional name for the tool (defaults to function name).
            description: Optional description of what the tool does.
            credits: Credits to redeem per call. Can be:
                - None: No credits charged (free tool)
                - int: Fixed number of credits
                - Callable[[args, result], int]: Dynamic credits based on args/result
            on_redeem_error: What to do if credit redemption fails.
                - "ignore": Silently ignore (default)
                - "propagate": Raise an error

        Returns:
            Decorator function.

        Example:
            >>> @mcp.tool(credits=5)
            >>> def weather(city: str) -> str:
            ...     '''Get weather for a city.
            ...
            ...     :param city: City name
            ...     '''
            ...     return f"Weather in {city}: 22°C"
        """
        if callable(name):
            raise TypeError(
                "The @tool decorator was used incorrectly. "
                "Did you forget to call it? Use @tool() instead of @tool"
            )

        def decorator(fn: F) -> F:
            tool_name = name or fn.__name__
            tool_description = description or fn.__doc__ or ""

            # Extract input schema from type hints
            input_schema = _extract_input_schema(fn)

            # Create handler that adapts function signature to MCP format
            @functools.wraps(fn)
            async def handler(args: Dict[str, Any], context: Any = None) -> Any:
                # Call the original function with unpacked args
                if asyncio.iscoroutinefunction(fn):
                    result = await fn(**args)
                else:
                    result = fn(**args)

                # Convert result to MCP format
                if isinstance(result, dict) and "content" in result:
                    return result
                elif isinstance(result, str):
                    return {"content": [{"type": "text", "text": result}]}
                else:
                    return {"content": [{"type": "text", "text": json.dumps(result)}]}

            # Build tool config
            tool_config = {
                "description": (
                    tool_description.split("\n")[0].strip() if tool_description else ""
                ),
                "inputSchema": input_schema,
            }

            # Build options
            options = {}
            if credits is not None:
                options["credits"] = credits
            if on_redeem_error != "ignore":
                options["onRedeemError"] = on_redeem_error

            # Register with server manager
            self._manager.register_tool(
                tool_name,
                tool_config,
                handler,
                options if options else None,
            )

            # Track for introspection
            self._registered_tools[tool_name] = {
                "name": tool_name,
                "description": tool_description,
                "credits": credits,
                "fn": fn,
            }

            return fn

        return decorator

    def resource(
        self,
        uri: str,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        mime_type: Optional[str] = None,
        credits: Optional[CreditsOption] = None,
        on_redeem_error: str = "ignore",
    ) -> Callable[[F], F]:
        """Decorator to register a resource with optional credit redemption.

        Args:
            uri: URI for the resource (e.g., "data://config" or "data://{id}").
            name: Optional name for the resource.
            description: Optional description of the resource.
            mime_type: Optional MIME type for the resource.
            credits: Credits to redeem per read.
            on_redeem_error: What to do if credit redemption fails.

        Returns:
            Decorator function.

        Example:
            >>> @mcp.resource("data://config", credits=2)
            >>> def get_config() -> str:
            ...     return '{"version": "1.0.0"}'
        """
        if callable(uri):
            raise TypeError(
                "The @resource decorator was used incorrectly. "
                "Did you forget to call it? Use @resource('uri') instead of @resource"
            )

        def decorator(fn: F) -> F:
            resource_name = name or fn.__name__
            resource_description = description or fn.__doc__ or ""

            # Create handler that adapts function signature to MCP format
            @functools.wraps(fn)
            async def handler(
                uri_obj: Any, variables: Dict[str, str], context: Any = None
            ) -> Any:
                # Get function parameters
                sig = inspect.signature(fn)
                params = list(sig.parameters.keys())

                # Call the original function
                if asyncio.iscoroutinefunction(fn):
                    if params and params[0] not in ("self", "cls"):
                        # Function expects parameters (template resource)
                        result = await fn(**variables)
                    else:
                        result = await fn()
                else:
                    if params and params[0] not in ("self", "cls"):
                        result = fn(**variables)
                    else:
                        result = fn()

                # Convert result to MCP format
                if isinstance(result, dict) and "contents" in result:
                    return result
                elif isinstance(result, str):
                    return {
                        "contents": [
                            {
                                "uri": str(uri_obj) if uri_obj else uri,
                                "mimeType": mime_type or "text/plain",
                                "text": result,
                            }
                        ]
                    }
                else:
                    return {
                        "contents": [
                            {
                                "uri": str(uri_obj) if uri_obj else uri,
                                "mimeType": mime_type or "application/json",
                                "text": json.dumps(result),
                            }
                        ]
                    }

            # Build resource config
            resource_config = {
                "name": resource_name,
                "description": (
                    resource_description.split("\n")[0].strip()
                    if resource_description
                    else ""
                ),
                "mimeType": mime_type or "text/plain",
            }

            # Build options
            options = {}
            if credits is not None:
                options["credits"] = credits
            if on_redeem_error != "ignore":
                options["onRedeemError"] = on_redeem_error

            # Register with server manager
            self._manager.register_resource(
                uri,
                resource_config,
                handler,
                options if options else None,
            )

            # Track for introspection
            self._registered_resources[uri] = {
                "uri": uri,
                "name": resource_name,
                "description": resource_description,
                "credits": credits,
                "fn": fn,
            }

            return fn

        return decorator

    def prompt(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        credits: Optional[CreditsOption] = None,
        on_redeem_error: str = "ignore",
    ) -> Callable[[F], F]:
        """Decorator to register a prompt with optional credit redemption.

        Args:
            name: Optional name for the prompt (defaults to function name).
            description: Optional description of what the prompt does.
            credits: Credits to redeem per use.
            on_redeem_error: What to do if credit redemption fails.

        Returns:
            Decorator function.

        Example:
            >>> @mcp.prompt(credits=1)
            >>> def greeting(name: str) -> list:
            ...     return [{"role": "user", "content": f"Hello {name}!"}]
        """
        if callable(name):
            raise TypeError(
                "The @prompt decorator was used incorrectly. "
                "Did you forget to call it? Use @prompt() instead of @prompt"
            )

        def decorator(fn: F) -> F:
            prompt_name = name or fn.__name__
            prompt_description = description or fn.__doc__ or ""

            # Extract arguments from function signature
            sig = inspect.signature(fn)
            arguments = []
            for param_name, param in sig.parameters.items():
                if param_name in ("self", "cls", "context", "ctx"):
                    continue
                arg_info: Dict[str, Any] = {"name": param_name}
                if param.default != inspect.Parameter.empty:
                    arg_info["required"] = False
                else:
                    arg_info["required"] = True
                arguments.append(arg_info)

            # Create handler that adapts function signature to MCP format
            @functools.wraps(fn)
            async def handler(args: Dict[str, Any], context: Any = None) -> Any:
                # Call the original function with unpacked args
                if asyncio.iscoroutinefunction(fn):
                    result = await fn(**args)
                else:
                    result = fn(**args)

                # Convert result to MCP format
                if isinstance(result, dict) and "messages" in result:
                    return result
                elif isinstance(result, list):
                    return {"messages": result}
                else:
                    return {
                        "messages": [
                            {
                                "role": "user",
                                "content": {"type": "text", "text": str(result)},
                            }
                        ]
                    }

            # Build prompt config
            prompt_config = {
                "name": prompt_name,
                "description": (
                    prompt_description.split("\n")[0].strip()
                    if prompt_description
                    else ""
                ),
                "arguments": arguments,
            }

            # Build options
            options = {}
            if credits is not None:
                options["credits"] = credits
            if on_redeem_error != "ignore":
                options["onRedeemError"] = on_redeem_error

            # Register with server manager
            self._manager.register_prompt(
                prompt_name,
                prompt_config,
                handler,
                options if options else None,
            )

            # Track for introspection
            self._registered_prompts[prompt_name] = {
                "name": prompt_name,
                "description": prompt_description,
                "credits": credits,
                "fn": fn,
            }

            return fn

        return decorator

    async def start(
        self,
        port: int,
        host: str = "0.0.0.0",
        base_url: Optional[str] = None,
        on_log: Optional[Callable[[str], None]] = None,
    ) -> Dict[str, Any]:
        """Start the MCP server with OAuth and paywall.

        This starts a complete MCP server with:
        - OAuth 2.1 discovery endpoints (/.well-known/*)
        - Client registration (/register)
        - Health check (/health)
        - Server info (/)
        - MCP handlers (POST/GET/DELETE /mcp)
        - Credit redemption for protected handlers

        Args:
            port: Port to listen on.
            host: Host to bind to (default: "0.0.0.0").
            base_url: Base URL for the server (default: http://localhost:{port}).
            on_log: Optional callback for logging.

        Returns:
            Dict with:
                - info: Server info (baseUrl, port, tools, resources, prompts)
                - stop: Async function to stop the server

        Example:
            >>> result = await mcp.start(port=3002)
            >>> print(f"Server running at {result['info']['baseUrl']}")
            >>> # Later...
            >>> await result["stop"]()
        """
        if not self.agent_id:
            raise ValueError("agent_id is required. Set it in the constructor.")

        config = {
            "port": port,
            "host": host,
            "agentId": self.agent_id,
            "serverName": self.name,
            "version": self.version,
            "description": self.description,
            "baseUrl": base_url,
            "onLog": on_log or (lambda msg: print(f"[MCP] {msg}")),
        }

        return await self._manager.start(config)

    async def stop(self) -> None:
        """Stop the MCP server gracefully.

        Example:
            >>> await mcp.stop()
        """
        await self._manager.stop()

    # =========================================================================
    # Introspection methods
    # =========================================================================

    def list_tools(self) -> List[str]:
        """List all registered tool names.

        Returns:
            List of tool names.
        """
        return list(self._registered_tools.keys())

    def list_resources(self) -> List[str]:
        """List all registered resource URIs.

        Returns:
            List of resource URIs.
        """
        return list(self._registered_resources.keys())

    def list_prompts(self) -> List[str]:
        """List all registered prompt names.

        Returns:
            List of prompt names.
        """
        return list(self._registered_prompts.keys())

    def get_tool_info(self, name: str) -> Optional[Dict[str, Any]]:
        """Get information about a registered tool.

        Args:
            name: Tool name.

        Returns:
            Tool info dict or None if not found.
        """
        return self._registered_tools.get(name)

    def get_resource_info(self, uri: str) -> Optional[Dict[str, Any]]:
        """Get information about a registered resource.

        Args:
            uri: Resource URI.

        Returns:
            Resource info dict or None if not found.
        """
        return self._registered_resources.get(uri)

    def get_prompt_info(self, name: str) -> Optional[Dict[str, Any]]:
        """Get information about a registered prompt.

        Args:
            name: Prompt name.

        Returns:
            Prompt info dict or None if not found.
        """
        return self._registered_prompts.get(name)



"""
Strands agent tool decorator for Nevermined payment protection using the x402 protocol.

This module provides a decorator to protect Strands agent tools with
Nevermined payment verification and settlement following the x402 protocol.

Example usage:
    ```python
    from strands import tool, Agent
    from payments_py import Payments, PaymentOptions
    from payments_py.x402.strands import requires_payment, PaymentContext

    payments = Payments.get_instance(
        PaymentOptions(nvm_api_key="...", environment="sandbox")
    )

    @tool
    @requires_payment(payments=payments, plan_id="plan-123", credits=1)
    def analyze_data(query: str, tool_context=None) -> dict:
        \"\"\"Analyze data with payment protection.

        Args:
            query: The analysis query
        \"\"\"
        return {"status": "success", "content": [{"text": f"Analysis: {query}"}]}

    # Create agent with payment-protected tools
    agent = Agent(tools=[analyze_data])

    # Invoke with payment token via invocation_state
    result = agent("Analyze sales data", payment_token="x402-access-token")
    ```

For full documentation, see the decorator module.
"""

from payments_py.x402.types import PaymentContext
from .decorator import (
    requires_payment,
    extract_payment_required,
    CreditsCallable,
)

__all__ = [
    "requires_payment",
    "extract_payment_required",
    "PaymentContext",
    "CreditsCallable",
]



"""
The PlansAPI class provides methods to register and interact with payment plans on Nevermined.
"""

import requests
from typing import Dict, Any, Optional, Literal
from payments_py.common.payments_error import PaymentsError
from payments_py.common.types import (
    PaymentOptions,
    PlanMetadata,
    PlanPriceConfig,
    PlanCreditsConfig,
    PlanRedemptionType,
    PaginationOptions,
    PlanBalance,
)
from payments_py.api.base_payments import BasePaymentsAPI
from payments_py.api.nvm_api import (
    API_URL_REGISTER_PLAN,
    API_URL_GET_PLAN,
    API_URL_PLAN_BALANCE,
    API_URL_ORDER_PLAN,
    API_URL_MINT_PLAN,
    API_URL_MINT_EXPIRABLE_PLAN,
    API_URL_BURN_PLAN,
    API_URL_GET_PLAN_AGENTS,
    API_URL_STRIPE_CHECKOUT,
    API_URL_REDEEM_PLAN,
)
from payments_py.utils import get_random_big_int, is_ethereum_address
from payments_py import plans as plan_utils
from payments_py.plans import (
    get_pay_as_you_go_price_config,
    get_pay_as_you_go_credits_config,
)
from payments_py.api.contracts_api import ContractsAPI


class PlansAPI(BasePaymentsAPI):
    """
    The PlansAPI class provides methods to register and interact with payment plans on Nevermined.
    """

    def __init__(self, options: PaymentOptions):
        """
        Initialize the PlansAPI class.

        Args:
            options: The options to initialize the payments class
        """
        super().__init__(options)
        self.contracts_api = ContractsAPI(options)

    @classmethod
    def get_instance(cls, options: PaymentOptions) -> "PlansAPI":
        """
        Get a singleton instance of the PlansAPI class.

        Args:
            options: The options to initialize the payments class

        Returns:
            The instance of the PlansAPI class
        """
        return cls(options)

    def register_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
        nonce: Optional[int] = None,
        access_limit: Optional[
            Literal["credits", "time"]
        ] = None,  # 'credits' or 'time'
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Payment Plan on Nevermined in a flexible manner.
        A Nevermined Credits Plan limits access based on plan usage.
        With them, AI Builders control the number of requests that can be made to an agent or service.
        Every time a user accesses any resource associated with the Payment Plan, the usage consumes from a capped amount of credits.
        When the user consumes all the credits, the plan automatically expires and the user needs to top up to continue using the service.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration
            nonce: Optional nonce for the transaction

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan

        Raises:
            PaymentsError: If registration fails
        """
        if access_limit and access_limit not in ["credits", "time"]:
            raise PaymentsError.validation(
                "Invalid access limit",
                "accessLimit must be either 'credits' or 'time'",
            )
        if not access_limit:
            access_limit = "time" if credits_config.duration_secs > 0 else "credits"

        if nonce is None:
            nonce = get_random_big_int()

        body = {
            "metadataAttributes": self.pydantic_to_dict(plan_metadata),
            "priceConfig": self.pydantic_to_dict(price_config),
            "creditsConfig": self.pydantic_to_dict(credits_config),
            "nonce": nonce,
            "isTrialPlan": getattr(plan_metadata, "is_trial_plan", False),
            "accessLimit": access_limit,
        }

        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_REGISTER_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to register plan. {response.status_code} - {response.text}"
            )

        return response.json()

    def register_credits_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Payment Plan on Nevermined based on Credits.
        A Nevermined Credits Plan limits the access by the access/usage of the Plan.
        With them, AI Builders control the number of requests that can be made to an agent or service.
        Every time a user accesses any resource associated with the Payment Plan, the usage consumes from a capped amount of credits.
        When the user consumes all the credits, the plan automatically expires and the user needs to top up to continue using the service.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan

        Raises:
            PaymentsError: If the credits configuration is invalid
        """

        if credits_config.min_amount > credits_config.max_amount:
            raise PaymentsError.validation(
                "The creditsConfig.minAmount can not be more than creditsConfig.maxAmount"
            )

        return self.register_plan(
            plan_metadata, price_config, credits_config, access_limit="credits"
        )

    def register_time_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Payment Plan on Nevermined limited by duration.
        A Nevermined Credits Plan limits the access by the access/usage of the Plan.
        With them, AI Builders control the number of requests that can be made to an agent or service.
        Every time a user accesses any resource associated with the Payment Plan, the usage consumes from a capped amount of credits.
        When the user consumes all the credits, the plan automatically expires and the user needs to top up to continue using the service.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan

        Raises:
            PaymentsError: If the credits configuration is invalid
        """

        return self.register_plan(
            plan_metadata, price_config, credits_config, access_limit="time"
        )

    def register_credits_trial_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Trial Payment Plan on Nevermined based on Credits.
        A Nevermined Trial Plan allow subscribers of that plan to test the Agents associated to it.
        A Trial plan is a plan that only can be purchased once by a user.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan
        """
        plan_metadata.is_trial_plan = True
        return self.register_credits_plan(plan_metadata, price_config, credits_config)

    def register_time_trial_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Trial Payment Plan on Nevermined limited by duration.
        A Nevermined Trial Plan allow subscribers of that plan to test the Agents associated to it.
        A Trial plan is a plan that only can be purchased once by a user.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan
        """
        plan_metadata.is_trial_plan = True
        return self.register_time_plan(plan_metadata, price_config, credits_config)

    def get_plan(self, plan_id: str) -> Dict[str, Any]:
        """
        Get the metadata for a given Plan identifier.

        Args:
            plan_id: The unique identifier of the plan

        Returns:
            The plan's metadata

        Raises:
            PaymentsError: If the plan is not found
        """
        url = f"{self.environment.backend}{API_URL_GET_PLAN.format(plan_id=plan_id)}"
        response = requests.get(url)
        if not response.ok:
            raise PaymentsError.validation(
                f"Plan not found. {response.status_code} - {response.text}"
            )
        return response.json()

    def get_plan_balance(
        self, plan_id: str, account_address: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Get the balance of a plan for a specific account.

        Args:
            plan_id: The unique identifier of the plan
            account_address: The account address to check balance for (defaults to current user)

        Returns:
            The plan balance information with properly typed fields (balance as int)

        Raises:
            PaymentsError: If unable to get plan balance
        """

        if not is_ethereum_address(account_address):
            account_address = self.get_account_address()

        url = f"{self.environment.backend}{API_URL_PLAN_BALANCE.format(plan_id=plan_id, holder_address=account_address)}"
        response = requests.get(
            url,
            headers={"Accept": "application/json", "Content-Type": "application/json"},
        )
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to get plan balance. {response.status_code} - {response.text}"
            )

        # Parse and validate response using Pydantic model to ensure type conversion
        response_data = response.json()
        return PlanBalance(**response_data)

    def order_plan(self, plan_id: str) -> Dict[str, bool]:
        """
        Order a plan by its ID.

        Args:
            plan_id: The ID of the plan to order

        Returns:
            The result of the order operation

        Raises:
            PaymentsError: If unable to order the plan
        """
        options = self.get_backend_http_options("POST")
        url = f"{self.environment.backend}{API_URL_ORDER_PLAN}".format(plan_id=plan_id)

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to order plan. {response.status_code} - {response.text}"
            )
        return response.json()

    def mint_plan_credits(
        self, plan_id: str, credits_amount: int, credits_receiver: str
    ) -> Dict[str, Any]:
        """
        Mint credits for a plan.

        Args:
            plan_id: The unique identifier of the plan
            credits_amount: The amount of credits to mint
            credits_receiver: The address that will receive the credits

        Returns:
            The result of the mint operation

        Raises:
            PaymentsError: If unable to mint credits
        """
        body = {
            "planId": plan_id,
            "amount": credits_amount,
            "creditsReceiver": credits_receiver,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_MINT_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to mint credits. {response.status_code} - {response.text}"
            )
        return response.json()

    def mint_plan_expirable(
        self,
        plan_id: str,
        credits_amount: int,
        credits_receiver: str,
        credits_duration: int = 0,
    ) -> Dict[str, Any]:
        """
        Mint expirable credits for a plan.

        Args:
            plan_id: The unique identifier of the plan
            credits_amount: The amount of credits to mint
            credits_receiver: The address that will receive the credits
            credits_duration: The duration of the credits in seconds

        Returns:
            The result of the mint operation

        Raises:
            PaymentsError: If unable to mint credits
        """
        body = {
            "planId": plan_id,
            "creditsAmount": credits_amount,
            "creditsReceiver": credits_receiver,
            "creditsDuration": credits_duration,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_MINT_EXPIRABLE_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to mint expirable credits. {response.status_code} - {response.text}"
            )
        return response.json()

    def burn_credits(self, plan_id: str, credits_amount: str) -> Dict[str, Any]:
        """
        Burn credits from a plan.

        Args:
            plan_id: The unique identifier of the plan
            credits_amount: The amount of credits to burn

        Returns:
            The result of the burn operation

        Raises:
            PaymentsError: If unable to burn credits
        """
        body = {
            "planId": plan_id,
            "creditsAmount": credits_amount,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_BURN_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to burn credits. {response.status_code} - {response.text}"
            )
        return response.json()

    def get_agents_associated_to_plan(
        self, plan_id: str, pagination: Optional[PaginationOptions] = None
    ) -> Dict[str, Any]:
        """
        Gets the list of agents that can be accessed with a plan.

        Args:
            plan_id: The unique identifier of the plan
            pagination: Optional pagination options to control the number of results returned

        Returns:
            The list of all different agents giving access to the plan

        Raises:
            PaymentsError: If the plan is not found
        """
        if pagination is None:
            pagination = PaginationOptions()

        url = f"{self.environment.backend}{API_URL_GET_PLAN_AGENTS.format(plan_id=plan_id)}"
        params = {
            "page": pagination.page,
            "offset": pagination.offset,
        }
        response = requests.get(url, params=params)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to get agents associated to plan. {response.status_code} - {response.text}"
            )
        return response.json()

    # ------------------------------------------------------------------
    # Static helper methods to build price/credits configurations
    # These mirror functions in payments_py/plans.py for ergonomic access as
    # Payments.plans.<method>(...)
    # ------------------------------------------------------------------

    # Expose duration constants for convenience
    ONE_DAY_DURATION: int = plan_utils.ONE_DAY_DURATION
    ONE_WEEK_DURATION: int = plan_utils.ONE_WEEK_DURATION
    ONE_MONTH_DURATION: int = plan_utils.ONE_MONTH_DURATION
    ONE_YEAR_DURATION: int = plan_utils.ONE_YEAR_DURATION

    # Price configuration builders -------------------------------------
    @staticmethod
    def get_fiat_price_config(amount: int, receiver: str) -> PlanPriceConfig:
        """Build a fiat price configuration."""
        return plan_utils.get_fiat_price_config(amount, receiver)

    @staticmethod
    def get_crypto_price_config(
        amount: int,
        receiver: str,
        token_address: str = "0x0000000000000000000000000000000000000000",
    ) -> PlanPriceConfig:
        """Build a crypto (native/ERC20) price configuration."""
        return plan_utils.get_crypto_price_config(amount, receiver, token_address)

    @staticmethod
    def get_erc20_price_config(
        amount: int, token_address: str, receiver: str
    ) -> PlanPriceConfig:
        """Build an ERC20 price configuration."""
        return plan_utils.get_erc20_price_config(amount, token_address, receiver)

    @staticmethod
    def get_free_price_config() -> PlanPriceConfig:
        """Build a free price configuration."""
        return plan_utils.get_free_price_config()

    @staticmethod
    def get_native_token_price_config(amount: int, receiver: str) -> PlanPriceConfig:
        """Build a native token price configuration."""
        return plan_utils.get_native_token_price_config(amount, receiver)

    # Credits configuration builders -----------------------------------
    @staticmethod
    def get_expirable_duration_config(duration_of_plan: int) -> PlanCreditsConfig:
        """Build an expirable duration credits configuration."""
        return plan_utils.get_expirable_duration_config(duration_of_plan)

    @staticmethod
    def get_non_expirable_duration_config() -> PlanCreditsConfig:
        """Build a non-expirable duration credits configuration."""
        return plan_utils.get_non_expirable_duration_config()

    @staticmethod
    def get_fixed_credits_config(
        credits_granted: int, credits_per_request: int = 1
    ) -> PlanCreditsConfig:
        """Build a fixed credits configuration."""
        return plan_utils.get_fixed_credits_config(credits_granted, credits_per_request)

    @staticmethod
    def get_dynamic_credits_config(
        credits_granted: int,
        min_credits_per_request: int = 1,
        max_credits_per_request: int = 1,
    ) -> PlanCreditsConfig:
        """Build a dynamic credits configuration."""
        return plan_utils.get_dynamic_credits_config(
            credits_granted, min_credits_per_request, max_credits_per_request
        )

    @staticmethod
    def set_redemption_type(
        credits_config: PlanCreditsConfig, redemption_type: PlanRedemptionType
    ) -> PlanCreditsConfig:
        """Set redemption type on a credits configuration (returns new object)."""
        return plan_utils.set_redemption_type(credits_config, redemption_type)

    @staticmethod
    def set_proof_required(
        credits_config: PlanCreditsConfig, proof_required: bool = True
    ) -> PlanCreditsConfig:
        """Set proof requirement on a credits configuration (returns new object)."""
        return plan_utils.set_proof_required(credits_config, proof_required)

    # Pay As You Go configuration builders -----------------------------------
    def get_pay_as_you_go_price_config(
        self,
        amount: int,
        receiver: str,
        token_address: str = plan_utils.ZeroAddress,
    ) -> PlanPriceConfig:
        """
        Build a pay-as-you-go price configuration using contract address from API.

        This method fetches the PayAsYouGoTemplate contract address from the API info endpoint
        and uses it to create the price configuration. The address is cached for subsequent calls.

        Args:
            amount: The amount per usage in the smallest unit of the token
            receiver: The address that will receive the payment
            token_address: The address of the token to use for payment (defaults to native token)

        Returns:
            A PlanPriceConfig object configured for pay-as-you-go payments

        Raises:
            PaymentsError: If unable to fetch contract address from API
            ValueError: If the receiver address is not a valid Ethereum address
        """
        # Get contract address from contracts API
        template_address = self.contracts_api.contracts.pay_as_you_go_template

        return get_pay_as_you_go_price_config(
            amount, receiver, token_address, template_address=template_address
        )

    @staticmethod
    def get_pay_as_you_go_credits_config() -> PlanCreditsConfig:
        """Build a pay-as-you-go credits configuration."""
        return get_pay_as_you_go_credits_config()

    def order_fiat_plan(self, plan_id: str) -> Dict[str, Any]:
        """
        Order a fiat plan using Stripe checkout.

        Args:
            plan_id: The unique identifier of the plan

        Returns:
            The Stripe checkout result

        Raises:
            PaymentsError: If unable to order the fiat plan
        """
        body = {"planId": plan_id, "sessionType": "embedded"}
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_STRIPE_CHECKOUT}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to order fiat plan. {response.status_code} - {response.text}"
            )
        return response.json()

    def redeem_credits(
        self,
        agent_request_id: str,
        plan_id: str,
        redeem_from: str,
        credits_amount_to_redeem: str,
    ) -> Dict[str, Any]:
        """
        Redeem credits from a plan for a specific agent request.

        Args:
            agent_request_id: The unique identifier of the agent request
            plan_id: The unique identifier of the plan
            redeem_from: The address to redeem credits from
            credits_amount_to_redeem: The amount of credits to redeem

        Returns:
            The result of the redeem operation

        Raises:
            PaymentsError: If unable to redeem credits
        """
        body = {
            "agentRequestId": agent_request_id,
            "planId": plan_id,
            "redeemFrom": redeem_from,
            "amount": credits_amount_to_redeem,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_REDEEM_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to redeem credits. {response.status_code} - {response.text}"
            )
        return response.json()



"""
Server helpers for declaring Nevermined extension metadata.

These functions help resource servers attach Nevermined payment requirements
to their PaymentRequired responses following the x402 v2 extension pattern.
"""

from typing import Optional

from .types import NeverminedExtension, NeverminedInfo


def declare_nevermined_extension(
    plan_id: str,
    agent_id: str,
    max_amount: str,
    network: str = "base-sepolia",
    scheme: str = "contract",
    environment: Optional[str] = None,
    subscriber_address: Optional[str] = None,
) -> NeverminedExtension:
    """
    Create a Nevermined extension for a PaymentRequired response.

    This helper creates properly formatted Nevermined extension metadata
    that can be included in a PaymentRequired response. It follows the
    x402 v2 extension pattern of info + schema.

    Args:
        plan_id: Nevermined pricing plan ID
        agent_id: Nevermined AI agent ID
        max_amount: Maximum credits to burn per request (as string)
        network: Blockchain network (default: "base-sepolia")
        scheme: Payment scheme (default: "contract")
        environment: Nevermined environment ("sandbox", "live")
        subscriber_address: Optional subscriber blockchain address

    Returns:
        NeverminedExtension with info and schema following x402 v2 pattern

    Example:
        >>> from payments_py.x402.extensions.nevermined import (
        ...     declare_nevermined_extension,
        ...     NEVERMINED
        ... )
        >>> from payments_py.x402.types_v2 import (
        ...     PaymentRequiredResponseV2,
        ...     ResourceInfo
        ... )
        >>>
        >>> extension = declare_nevermined_extension(
        ...     plan_id="85917684554499762134516240562181895926019634254204202319880150802501990701934",
        ...     agent_id="80918427023170428029540261117198154464497879145267720259488529685089104529015",
        ...     max_amount="2",
        ...     network="base-sepolia",
        ...     environment="sandbox"
        ... )
        >>>
        >>> payment_required = PaymentRequiredResponseV2(
        ...     x402_version=2,
        ...     resource=ResourceInfo(url="https://api.example.com/data"),
        ...     accepts=[...],
        ...     extensions={
        ...         NEVERMINED: extension
        ...     }
        ... )
    """
    # Build the info object with required fields
    info: NeverminedInfo = {
        "plan_id": plan_id,
        "agent_id": agent_id,
        "max_amount": max_amount,
        "network": network,
        "scheme": scheme,
    }

    # Add optional fields if provided
    if environment:
        info["environment"] = environment
    if subscriber_address:
        info["subscriber_address"] = subscriber_address

    # Define JSON Schema for validation
    # This allows facilitators to validate the extension data
    schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "properties": {
            "plan_id": {
                "type": "string",
                "description": "Nevermined pricing plan ID",
            },
            "agent_id": {
                "type": "string",
                "description": "Nevermined AI agent ID",
            },
            "max_amount": {
                "type": "string",
                "description": "Maximum credits to burn per request",
            },
            "network": {
                "type": "string",
                "description": "Blockchain network (e.g., base-sepolia)",
            },
            "scheme": {
                "type": "string",
                "description": "Payment scheme (e.g., contract)",
            },
            "environment": {
                "type": "string",
                "description": "Nevermined environment (staging_sandbox, staging_live, sandbox, live, custom)",
                "enum": [
                    "staging_sandbox",
                    "staging_live",
                    "sandbox",
                    "live",
                    "custom",
                ],
            },
            "subscriber_address": {
                "type": "string",
                "description": "Subscriber's blockchain address",
            },
        },
        "required": ["plan_id", "agent_id", "max_amount", "network", "scheme"],
        "additionalProperties": False,
    }

    return {"info": info, "schema": schema}


__all__ = ["declare_nevermined_extension"]



"""API classes for the Nevermined Payments Python SDK."""

from payments_py.api.agents_api import AgentsAPI
from payments_py.api.plans_api import PlansAPI
from payments_py.api.requests_api import AgentRequestsAPI
from payments_py.api.query_api import AIQueryApi
from payments_py.api.observability_api import ObservabilityAPI
from payments_py.api.contracts_api import ContractsAPI
from payments_py.api.base_payments import BasePaymentsAPI

__all__ = [
    "AgentsAPI",
    "PlansAPI",
    "AgentRequestsAPI",
    "AIQueryApi",
    "ObservabilityAPI",
    "ContractsAPI",
    "BasePaymentsAPI",
]



"""
The FacilitatorAPI class provides methods to verify and settle AI agent permissions using X402 access tokens.
This allows AI agents to act as facilitators, verifying and settling credits on behalf of subscribers.

Example usage:
    from payments_py import Payments, PaymentOptions
    from payments_py.x402 import X402PaymentRequired, X402Scheme

    # Initialize the Payments instance
    payments = Payments.get_instance(
        PaymentOptions(
            nvm_api_key="your-nvm-api-key",
            environment="sandbox"
        )
    )

    # The server's 402 PaymentRequired response
    payment_required = X402PaymentRequired(
        x402_version=2,
        accepts=[
            X402Scheme(
                scheme="nvm:erc4337",
                network="eip155:84532",
                plan_id="123",
            )
        ],
        extensions={}
    )

    # Get X402 access token from payment-signature header (x402 v2 spec)
    x402_token = request.headers.get("payment-signature")

    # Verify if subscriber has sufficient permissions/credits
    verification = payments.facilitator.verify_permissions(
        payment_required=payment_required,
        x402_access_token=x402_token,
        max_amount="2"  # optional
    )

    if verification.is_valid:
        # Settle (burn) the credits
        settlement = payments.facilitator.settle_permissions(
            payment_required=payment_required,
            x402_access_token=x402_token,
            max_amount="2"  # optional
        )
        print(f"Credits redeemed: {settlement.credits_redeemed}")
"""

import requests
from typing import Optional
from payments_py.common.payments_error import PaymentsError
from payments_py.common.types import PaymentOptions
from payments_py.api.base_payments import BasePaymentsAPI
from payments_py.api.nvm_api import (
    API_URL_VERIFY_PERMISSIONS,
    API_URL_SETTLE_PERMISSIONS,
)
from payments_py.x402.types import VerifyResponse, SettleResponse, X402PaymentRequired


class FacilitatorAPI(BasePaymentsAPI):
    """
    The FacilitatorAPI class provides methods to verify and settle AI agent permissions.
    It enables AI agents to act as facilitators, managing credit verification and settlement
    for subscribers using X402 access tokens.
    """

    @classmethod
    def get_instance(cls, options: PaymentOptions) -> "FacilitatorAPI":
        """
        Get a singleton instance of the FacilitatorAPI class.

        Args:
            options: The options to initialize the payments class

        Returns:
            The instance of the FacilitatorAPI class
        """
        return cls(options)

    def verify_permissions(
        self,
        payment_required: X402PaymentRequired,
        x402_access_token: str,
        max_amount: Optional[str] = None,
    ) -> VerifyResponse:
        """
        Verify if a subscriber has permission to use credits from a payment plan.
        This method simulates the credit usage without actually burning credits,
        checking if the subscriber has sufficient balance and permissions.

        The planId and subscriberAddress are extracted from the x402AccessToken.

        Args:
            payment_required: x402 PaymentRequired from 402 response (required, for validation)
            x402_access_token: The X402 access token (contains planId, subscriberAddress, agentId)
            max_amount: The maximum number of credits to verify (as string, optional)

        Returns:
            VerifyResponse with is_valid boolean and optional error details

        Raises:
            PaymentsError: If verification fails
        """
        url = f"{self.environment.backend}{API_URL_VERIFY_PERMISSIONS}"

        body: dict = {
            "paymentRequired": payment_required.model_dump(by_alias=True),
            "x402AccessToken": x402_access_token,
        }

        if max_amount is not None:
            body["maxAmount"] = max_amount

        options = self.get_public_http_options("POST", body)

        try:
            response = requests.post(url, **options)
            response.raise_for_status()
            return VerifyResponse.model_validate(response.json())
        except requests.HTTPError as err:
            try:
                error_message = response.json().get(
                    "message", "Permission verification failed"
                )
            except Exception:
                error_message = "Permission verification failed"
            raise PaymentsError.from_backend(
                error_message,
                {"code": f"HTTP {response.status_code}"},
            ) from err
        except Exception as err:
            if isinstance(err, PaymentsError):
                raise
            raise PaymentsError.from_backend(
                "Network error during permission verification",
                {"code": "network_error", "message": str(err)},
            ) from err

    def settle_permissions(
        self,
        payment_required: X402PaymentRequired,
        x402_access_token: str,
        max_amount: Optional[str] = None,
        agent_request_id: Optional[str] = None,
    ) -> SettleResponse:
        """
        Settle (burn) credits from a subscriber's payment plan.
        This method executes the actual credit consumption, burning the specified
        number of credits from the subscriber's balance. If the subscriber doesn't
        have enough credits, it will attempt to order more before settling.

        The planId and subscriberAddress are extracted from the x402AccessToken.

        Args:
            payment_required: x402 PaymentRequired from 402 response (required, for validation)
            x402_access_token: The X402 access token (contains planId, subscriberAddress, agentId)
            max_amount: The number of credits to burn (as string, optional)
            agent_request_id: Agent request ID for observability tracking (optional)

        Returns:
            SettleResponse with success boolean and transaction details

        Raises:
            PaymentsError: If settlement fails
        """
        url = f"{self.environment.backend}{API_URL_SETTLE_PERMISSIONS}"

        body: dict = {
            "paymentRequired": payment_required.model_dump(by_alias=True),
            "x402AccessToken": x402_access_token,
        }

        if max_amount is not None:
            body["maxAmount"] = max_amount

        if agent_request_id is not None:
            body["agentRequestId"] = agent_request_id

        options = self.get_public_http_options("POST", body)

        try:
            response = requests.post(url, **options)
            response.raise_for_status()
            return SettleResponse.model_validate(response.json())
        except requests.HTTPError as err:
            try:
                error_message = response.json().get(
                    "message", "Permission settlement failed"
                )
            except Exception:
                error_message = "Permission settlement failed"
            raise PaymentsError.from_backend(
                error_message,
                {"code": f"HTTP {response.status_code}"},
            ) from err
        except Exception as err:
            if isinstance(err, PaymentsError):
                raise
            raise PaymentsError.from_backend(
                "Network error during permission settlement",
                {"code": "network_error", "message": str(err)},
            ) from err



import os
from dataclasses import dataclass
from typing import Literal


@dataclass
class EnvironmentInfo:
    """
    Data class to store environment information.

    Attributes:
        frontend (str): Frontend URL
        backend (str): Backend URL
        proxy (str): Proxy URL
        helicone_url (str): Helicone URL
    """

    backend: str
    proxy: str
    helicone_url: str
    frontend: str = None


# Zero address constant
ZeroAddress = "0x0000000000000000000000000000000000000000"

# Supported environment names
EnvironmentName = Literal[
    "sandbox",
    "live",
    "staging_sandbox",
    "staging_live",
    "custom",
]

# Environments dictionary
Environments = {
    "staging_sandbox": EnvironmentInfo(
        frontend="https://nevermined.dev",
        backend="https://api.sandbox.nevermined.dev",
        proxy="https://proxy.sandbox.nevermined.dev",
        helicone_url="https://helicone.nevermined.dev",
    ),
    "staging_live": EnvironmentInfo(
        frontend="https://nevermined.dev",
        backend="https://api.live.nevermined.dev",
        proxy="https://proxy.live.nevermined.dev",
        helicone_url="https://helicone.nevermined.dev",
    ),
    "sandbox": EnvironmentInfo(
        frontend="https://nevermined.app",
        backend="https://api.sandbox.nevermined.app",
        proxy="https://proxy.sandbox.nevermined.app",
        helicone_url="https://helicone.nevermined.dev",
    ),
    "live": EnvironmentInfo(
        frontend="https://nevermined.app",
        backend="https://api.live.nevermined.app",
        proxy="https://proxy.live.nevermined.app",
        helicone_url="https://helicone.nevermined.dev",
    ),
    "custom": EnvironmentInfo(
        frontend=os.getenv("NVM_FRONTEND_URL", "http://localhost:3000"),
        backend=os.getenv("NVM_BACKEND_URL", "http://localhost:3001"),
        proxy=os.getenv("NVM_PROXY_URL", "https://localhost:443"),
        helicone_url=os.getenv("HELICONE_URL", "http://localhost:8585"),
    ),
}


def get_environment(name: EnvironmentName) -> EnvironmentInfo:
    """
    Get the environment configuration by name.

    Args:
        name: The name of the environment.

    Returns:
        EnvironmentInfo: The environment configuration.

    Raises:
        ValueError: If the environment name is not defined.
    """
    if name not in Environments:
        raise ValueError(f"Environment '{name}' is not defined.")
    return Environments[name]



"""Helper to build an AgentCard enriched with payment metadata (Python version)."""

from __future__ import annotations

from typing import Any, Dict, List

from payments_py.a2a.types import PaymentAgentCardMetadata


def build_payment_agent_card(
    base_card: Dict[str, Any], payment_metadata: PaymentAgentCardMetadata
) -> Dict[str, Any]:  # noqa: D401
    """Return a new agent card with payments extension.

    Args:
        base_card: The original agent card (dict following a2a.types.AgentCard schema).
        payment_metadata: Dict with payment information.

    Raises:
        ValueError: If required fields are missing or invalid.

    Returns:
        A copy of *base_card* that contains the payment extension in
        ``capabilities.extensions``.
    """
    # ------------------------------------------------------------------
    # Basic validation (mirror TS)
    # ------------------------------------------------------------------
    if "paymentType" not in payment_metadata:
        raise ValueError("paymentType is required")

    credits = payment_metadata.get("credits", 0)
    if credits < 0:
        raise ValueError("credits cannot be negative")

    if payment_metadata.get("isTrialPlan"):
        # Trial plan can have 0 credits, nothing else to check
        pass
    else:
        if credits <= 0:
            raise ValueError("credits must be a positive number for paid plans")

    if not payment_metadata.get("agentId"):
        raise ValueError("agentId is required")

    # ------------------------------------------------------------------
    # Build new card
    # ------------------------------------------------------------------
    extensions: List[Dict[str, Any]] = (
        base_card.get("capabilities", {}).get("extensions", []) or []
    )

    payment_extension = {
        "uri": "urn:nevermined:payment",
        "description": payment_metadata.get("costDescription"),
        "required": False,
        "params": dict(payment_metadata),  # cast to plain dict
    }

    new_extensions = [*extensions, payment_extension]

    capabilities = {
        **(base_card.get("capabilities", {})),
        "extensions": new_extensions,
    }

    return {**base_card, "capabilities": capabilities}



"""
Session manager for MCP transports.

This module handles creation, retrieval, and cleanup of SSE server transport
instances. It also stores request context (headers) per session for authentication.

Examples:
    >>> from payments_py.mcp.http import SessionManager, create_session_manager
    >>> manager = create_session_manager()
    >>> manager.set_mcp_server(mcp_server_instance)
    >>> session_id = manager.generate_session_id()
    >>> transport = await manager.get_or_create_session(session_id)
"""

import uuid
from typing import Any, Callable, Dict, List, Optional

from typing_extensions import TypedDict

# =============================================================================
# TYPES
# =============================================================================


class RequestContext(TypedDict, total=False):
    """Request context stored per session.

    Contains HTTP headers and other request info needed for authentication.

    Attributes:
        headers: HTTP headers dict.
        method: HTTP method (GET, POST, etc.).
        url: Request URL.
        ip: Client IP address.
    """

    headers: Dict[str, Any]
    method: Optional[str]
    url: Optional[str]
    ip: Optional[str]


class SessionManagerConfig(TypedDict, total=False):
    """Configuration for session manager.

    Attributes:
        onSessionCreated: Callback when a session is created.
        onSessionDestroyed: Callback when a session is destroyed.
        log: Logger function.
    """

    onSessionCreated: Optional[Callable[[str], None]]
    onSessionDestroyed: Optional[Callable[[str], None]]
    log: Optional[Callable[[str], None]]


# =============================================================================
# TRANSPORT LAZY LOADING
# =============================================================================


_SSEServerTransport: Optional[Any] = None


async def _get_transport_class() -> Any:
    """Lazily load the MCP SDK SSE server transport.

    Returns:
        SSEServerTransport class from mcp.server.sse module.

    Raises:
        ImportError: If the mcp SDK is not installed.
    """
    global _SSEServerTransport

    if _SSEServerTransport is None:
        try:
            from mcp.server.sse import SseServerTransport

            _SSEServerTransport = SseServerTransport
        except ImportError as error:
            raise ImportError(
                "Failed to load mcp SDK. Make sure it is installed: "
                "pip install mcp or pip install modelcontextprotocol"
            ) from error

    return _SSEServerTransport


# =============================================================================
# SESSION MANAGER
# =============================================================================


class SessionManager:
    """Manages MCP transport sessions.

    This class handles the lifecycle of MCP transport sessions, including
    creation, retrieval, and destruction. It also stores request context
    per session for authentication purposes.

    Attributes:
        _sessions: Dict mapping session IDs to transport instances.
        _request_contexts: Dict mapping session IDs to request contexts.
        _mcp_server: Reference to the MCP server instance.
        _config: Session manager configuration.

    Examples:
        >>> manager = SessionManager()
        >>> manager.set_mcp_server(server)
        >>> session_id = manager.generate_session_id()
        >>> transport = await manager.get_or_create_session(session_id)
        >>> manager.destroy_session(session_id)
    """

    def __init__(self, config: Optional[Dict[str, Any]] = None) -> None:
        """Initialize the session manager.

        Args:
            config: Optional configuration dict (SessionManagerConfig).
        """
        self._sessions: Dict[str, Any] = {}
        self._request_contexts: Dict[str, RequestContext] = {}
        self._mcp_server: Optional[Any] = None
        self._config: Dict[str, Any] = config or {}

    def set_mcp_server(self, server: Any) -> None:
        """Set the MCP server that transports will connect to.

        Args:
            server: MCP server instance (from mcp.server import Server).
        """
        self._mcp_server = server

    def generate_session_id(self) -> str:
        """Generate a new session ID.

        Returns:
            A new UUID session identifier.

        Examples:
            >>> manager = SessionManager()
            >>> session_id = manager.generate_session_id()
            >>> len(session_id) > 0
            True
        """
        return str(uuid.uuid4())

    def has_session(self, session_id: str) -> bool:
        """Check if a session exists.

        Args:
            session_id: Session identifier.

        Returns:
            True if session exists, False otherwise.
        """
        return session_id in self._sessions

    def get_session(self, session_id: str) -> Optional[Any]:
        """Get an existing session's transport.

        Args:
            session_id: Session identifier.

        Returns:
            Transport instance or None if session doesn't exist.
        """
        return self._sessions.get(session_id)

    def set_request_context(self, session_id: str, context: RequestContext) -> None:
        """Store request context (headers, etc.) for a session.

        This is called when an HTTP request arrives, before dispatching to MCP.

        Args:
            session_id: Session identifier.
            context: Request context dict with headers, method, url, ip.
        """
        self._request_contexts[session_id] = context

    def get_request_context(self, session_id: str) -> Optional[RequestContext]:
        """Get the stored request context for a session.

        Args:
            session_id: Session identifier.

        Returns:
            Request context dict or None if not found.
        """
        return self._request_contexts.get(session_id)

    def get_current_request_context(self) -> Optional[RequestContext]:
        """Get request context from the current/most recent request.

        This is used by paywall handlers to access headers when
        the session ID is not directly available.

        Returns:
            The most recently stored request context, or None.
        """
        # Return the most recent request context if any exist
        if self._request_contexts:
            # Get the most recently added context
            return next(reversed(self._request_contexts.values()), None)
        return None

    async def get_or_create_session(self, session_id: str) -> Any:
        """Get or create a transport for a session.

        If the session exists, returns the existing transport.
        If not, creates a new SSE transport and connects it to the MCP server.

        Args:
            session_id: Session identifier.

        Returns:
            Transport instance.

        Raises:
            RuntimeError: If MCP server is not set.
            ImportError: If mcp SDK is not installed.

        Examples:
            >>> manager = SessionManager()
            >>> manager.set_mcp_server(server)
            >>> transport = await manager.get_or_create_session("session-123")
        """
        # Return existing session if found
        existing_session = self._sessions.get(session_id)
        if existing_session is not None:
            return existing_session

        # Validate MCP server is set
        if not self._mcp_server:
            raise RuntimeError("MCP server not set. Call set_mcp_server() first.")

        # For Python MCP SDK, we don't use the SSE transport for HTTP mode.
        # Instead, we process requests directly via memory streams in mcp_handler.py.
        # Sessions are tracked by ID only for context management.

        # Create a simple session object to track state
        session = {
            "id": session_id,
            "created_at": __import__("datetime").datetime.utcnow().isoformat(),
        }

        # Store session
        self._sessions[session_id] = session

        self._log(f"Created new session {session_id}")

        on_session_created = self._config.get("onSessionCreated")
        if on_session_created:
            on_session_created(session_id)

        return session

    def destroy_session(self, session_id: str) -> bool:
        """Destroy a session.

        Args:
            session_id: Session identifier.

        Returns:
            True if session was destroyed, False if session didn't exist.

        Examples:
            >>> manager.destroy_session("session-123")
            True
        """
        if session_id in self._sessions:
            transport = self._sessions.pop(session_id)
            self._request_contexts.pop(session_id, None)

            self._log(f"Destroyed session {session_id}")

            on_session_destroyed = self._config.get("onSessionDestroyed")
            if on_session_destroyed:
                on_session_destroyed(session_id)

            # Try to close the transport if it has a close method
            if transport:
                if hasattr(transport, "close") and callable(transport.close):
                    try:
                        transport.close()
                    except Exception:
                        # Ignore close errors
                        pass
                elif hasattr(transport, "aclose") and callable(transport.aclose):
                    try:
                        import asyncio

                        asyncio.create_task(transport.aclose())
                    except Exception:
                        # Ignore close errors
                        pass

            return True

        return False

    def get_active_sessions(self) -> List[str]:
        """Get all active session IDs.

        Returns:
            List of active session identifiers.

        Examples:
            >>> manager.get_active_sessions()
            ['session-123', 'session-456']
        """
        return list(self._sessions.keys())

    def get_session_count(self) -> int:
        """Get the number of active sessions.

        Returns:
            Number of active sessions.

        Examples:
            >>> manager.get_session_count()
            2
        """
        return len(self._sessions)

    def destroy_all_sessions(self) -> None:
        """Destroy all sessions.

        Examples:
            >>> manager.destroy_all_sessions()
            # All sessions are destroyed
        """
        session_ids = self.get_active_sessions()
        for session_id in session_ids:
            self.destroy_session(session_id)

        self._log(f"Destroyed all {len(session_ids)} sessions")

    def _log(self, message: str) -> None:
        """Internal logging helper (no-op if log not provided).

        Args:
            message: Log message.
        """
        log_fn = self._config.get("log")
        if log_fn:
            log_fn(message)


# =============================================================================
# FACTORY FUNCTION
# =============================================================================


def create_session_manager(config: Optional[Dict[str, Any]] = None) -> SessionManager:
    """Create a new session manager.

    Args:
        config: Optional configuration dict (SessionManagerConfig).

    Returns:
        New SessionManager instance.

    Examples:
        >>> manager = create_session_manager({"log": print})
        >>> manager.generate_session_id()
        'e4a2c8b0-...'
    """
    return SessionManager(config)
